from typing import Any, Dict, cast

import aiohttp

from pyhoo.types import ApiResponse


class GetTickerDataTask:

    _path: str
    _ticker: str
    _params: Dict[str, Any]

    _BASE_URL = "https://query2.finance.yahoo.com"

    def __init__(
        self,
        path: str,
        ticker: str,
        **params: Any,
    ) -> None:
        self._path = path
        self._ticker = ticker
        self._params = params

    async def run(self, session: aiohttp.ClientSession) -> Dict[str, ApiResponse]:
        response = await session.request("GET", url=self._url)
        data = cast(Dict[str, ApiResponse], await response.json())
        return data

    @property
    def _url(self) -> str:
        url = f"{self._BASE_URL}/{self._path}/{self._ticker}"
        if self._params:
            url = f"{url}?{self._format_params()}"
        return url

    def _format_params(self) -> str:
        return "&".join(f"{name}={value}" for name, value in self._params.items())



from typing import TypedDict


class ReportedValueDict(TypedDict):

    raw: float
    fmt: str


class FundamentalsRowDict(TypedDict):

    dataId: int
    asOfDate: str
    periodType: str
    reportedValue: ReportedValueDict
    currencyCode: str


class FundamentalsMetaDict(TypedDict):

    symbol: str
    type: str


class FundamentalsDataRowDict(FundamentalsMetaDict):

    dataId: int
    asOfDate: str
    periodType: str
    reportedValue: float
    currencyCode: str



from typing import Iterable, TypedDict


class OptionDict(TypedDict):

    contractSymbol: str
    strike: float
    currency: str
    lastPrice: float
    change: float
    percentChange: float
    ask: float
    contractSize: str
    expiration: int
    lastTradeDate: int
    impliedVolatility: float
    inTheMoney: bool
    volume: int
    bid: float
    openInterest: int


class OptionQuoteDictBase(TypedDict):

    language: str
    region: str
    quoteType: str
    quoteSourceName: str
    triggerable: bool
    currency: str
    sharesOutstanding: int
    bookValue: float
    fiftyDayAverage: float
    fiftyDayAverageChange: float
    fiftyDayAverageChangePercent: float
    twoHundredDayAverage: float
    twoHundredDayAverageChange: float
    twoHundredDayAverageChangePercent: float
    marketCap: int
    forwardPE: float
    priceToBook: float
    sourceInterval: int
    exchangeDataDelayedBy: int
    tradeable: bool
    ask: float
    bidSize: int
    askSize: int
    fullExchangeName: str
    financialCurrency: str
    regularMarketOpen: float
    averageDailyVolume3Month: int
    averageDailyVolume10Day: int
    fiftyTwoWeekLowChange: float
    fiftyTwoWeekLowChangePercent: float
    fiftyTwoWeekRange: str
    fiftyTwoWeekHighChange: float
    fiftyTwoWeekHighChangePercent: float
    fiftyTwoWeekLow: float
    fiftyTwoWeekHigh: float
    dividendDate: int
    earningsTimestamp: int
    earningsTimestampStart: int
    earningsTimestampEnd: int
    trailingAnnualDividendRate: float
    trailingPE: float
    trailingAnnualDividendYield: float
    epsTrailingTwelveMonths: float
    epsForward: float
    epsCurrentYear: float
    priceEpsCurrentYear: float
    exchange: str
    shortName: str
    longName: str
    messageBoardId: str
    exchangeTimezoneName: str
    exchangeTimezoneShortName: str
    gmtOffSetMilliseconds: int
    market: str
    esgPopulated: bool
    marketState: str
    firstTradeDateMilliseconds: int
    priceHint: int
    regularMarketChange: float
    regularMarketChangePercent: float
    regularMarketTime: float
    regularMarketPrice: float
    regularMarketDayHigh: float
    regularMarketDayRange: str
    regularMarketDayLow: float
    regularMarketVolume: int
    regularMarketPreviousClose: float
    bid: float
    displayName: str
    symbol: str


class OptionQuoteDict(OptionQuoteDictBase, total=False):
    preMarketChange: float
    preMarketChangePercent: float
    preMarketTime: float
    preMarketPrice: float


class OptionsDict(TypedDict):

    expirationDate: int
    hasMiniOptions: bool
    calls: Iterable[OptionDict]
    puts: Iterable[OptionDict]


class OptionsDataRecord(OptionDict):
    underlyingSymbol: str
    type: str



from typing import Any, Iterable


class CustomException(Exception):
    pass


class UnknownParameterError(CustomException):
    def __init__(self, param: str, valid_params: Iterable[str]) -> None:
        super().__init__(f"Expected one of {[param for param in valid_params]}, got '{param}'.")


class InvalidParameterValueError(CustomException):
    def __init__(self, param: str, value: Any, options: Iterable[Any]) -> None:
        super().__init__(f"'{param}' has invalid value '{value}', must be one of {[option for option in options]}.")


class InvalidParameterTypeError(CustomException):
    def __init__(self, param: str, param_type: Any, config_type: Any) -> None:
        super().__init__(f"'{param}' has invalid type '{param_type}', expected {config_type}.")


class InvalidParameterPrefixError(CustomException):
    def __init__(self, param: str, value: Any, prefixes: Iterable[str]) -> None:
        super().__init__(f"'{param}' has invalid prefix, it must starts with one of {prefixes}.")


class MissingParameterError(CustomException):
    def __init__(self, param: str) -> None:
        super().__init__(f"Missing parameter '{param}'.")


class ApiError(CustomException):
    def __init__(self, code: str, description: str) -> None:
        super().__init__(f"Got error '{code}' from Yahoo Finance API, reason was: '{description}'")



from dataclasses import dataclass
from typing import Iterable, Optional

from pyhoo.models.abc import BaseModel, OptionalFieldsModel
from pyhoo.types.options import OptionDict


@dataclass(frozen=True)
class Option(OptionalFieldsModel):

    contractSymbol: str
    strike: float
    currency: str
    lastPrice: float
    change: float
    percentChange: float
    ask: float
    contractSize: str
    expiration: int
    lastTradeDate: int
    impliedVolatility: float
    inTheMoney: bool
    volume: int
    bid: float
    openInterest: int


@dataclass(frozen=True)
class OptionQuote(OptionalFieldsModel):

    language: str
    region: str
    quoteType: str
    quoteSourceName: str
    triggerable: bool
    currency: str
    sharesOutstanding: int
    bookValue: float
    fiftyDayAverage: float
    fiftyDayAverageChange: float
    fiftyDayAverageChangePercent: float
    twoHundredDayAverage: float
    twoHundredDayAverageChange: float
    twoHundredDayAverageChangePercent: float
    marketCap: int
    forwardPE: float
    priceToBook: float
    sourceInterval: int
    exchangeDataDelayedBy: int
    tradeable: bool
    ask: float
    bidSize: int
    askSize: int
    fullExchangeName: str
    financialCurrency: str
    regularMarketOpen: float
    averageDailyVolume3Month: int
    averageDailyVolume10Day: int
    fiftyTwoWeekLowChange: float
    fiftyTwoWeekLowChangePercent: float
    fiftyTwoWeekRange: str
    fiftyTwoWeekHighChange: float
    fiftyTwoWeekHighChangePercent: float
    fiftyTwoWeekLow: float
    fiftyTwoWeekHigh: float
    earningsTimestamp: int
    earningsTimestampStart: int
    earningsTimestampEnd: int
    trailingAnnualDividendRate: float
    trailingPE: float
    trailingAnnualDividendYield: float
    epsTrailingTwelveMonths: float
    epsForward: float
    epsCurrentYear: float
    priceEpsCurrentYear: float
    exchange: str
    shortName: str
    longName: str
    messageBoardId: str
    exchangeTimezoneName: str
    exchangeTimezoneShortName: str
    gmtOffSetMilliseconds: int
    market: str
    esgPopulated: bool
    marketState: str
    firstTradeDateMilliseconds: int
    priceHint: int
    regularMarketChange: float
    regularMarketChangePercent: float
    regularMarketTime: float
    regularMarketPrice: float
    regularMarketDayHigh: float
    regularMarketDayRange: str
    regularMarketDayLow: float
    regularMarketVolume: int
    regularMarketPreviousClose: float
    bid: float
    displayName: str
    symbol: str
    preMarketChange: float
    preMarketChangePercent: float
    preMarketTime: float
    preMarketPrice: float
    dividendDate: int


class Options(BaseModel):

    expirationDate: Optional[int]
    hasMiniOptions: Optional[bool]
    calls: Iterable[Option]
    puts: Iterable[Option]

    def __init__(
        self,
        expirationDate: Optional[int] = None,
        hasMiniOptions: Optional[bool] = None,
        calls: Optional[Iterable[OptionDict]] = None,
        puts: Optional[Iterable[OptionDict]] = None,
    ) -> None:
        self.expirationDate = expirationDate
        self.hasMiniOptions = hasMiniOptions
        self.calls = [Option(**call) for call in calls or []]
        self.puts = [Option(**put) for put in puts or []]



import abc
from dataclasses import dataclass
from typing import Any, Optional


class BaseModel(metaclass=abc.ABCMeta):
    """Base class for API response sub keys.
    Provide a user friendly `__repr__` method.
    """

    def __repr__(self) -> str:
        formatted_attrs = ", ".join(
            attr_name + "=" + attr_value.__repr__() for attr_name, attr_value in self.__dict__.items()
        )
        return f"{self.__class__.__name__}({formatted_attrs})"


@dataclass(frozen=True)
class OptionalFieldsModel(BaseModel):
    """For pure dataclasses without custom `__init__` method.
    Make all fields optional and default to `None`.
    It allows for less verbose dataclass declaration.
    """

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        for field, value in cls.__annotations__.items():
            cls.__annotations__[field] = Optional[value]
            if not hasattr(cls, field):
                setattr(cls, field, None)
        super().__init_subclass__(*args, **kwargs)  # type: ignore



from __future__ import annotations

import enum
from dataclasses import dataclass, field
from typing import List, Optional

from pyhoo.models.abc import BaseModel, OptionalFieldsModel
from pyhoo.types.chart import (
    ChartMetaDictBase,
    CurrentTradingPeriodDict,
    IndicatorsDict,
    TradingPeriodDict,
)


class Interval(enum.Enum):

    ONE_MIN = "1m"
    TWO_MIN = "2m"
    FIVE_MIN = "5m"
    FIFTEEN_MIN = "15m"
    THIRTY_MIN = "30m"
    ONE_HOUR = "1h"
    ONE_DAY = "1d"
    FIVE_DAYS = "5d"
    ONE_WEEK = "1w"
    ONE_MONTH = "1mo"
    THREE_MONTHS = "3mo"


@dataclass(frozen=True)
class Quote(OptionalFieldsModel):

    high: List[float]
    volume: List[float]
    open: List[float]
    close: List[float]
    low: List[float]


@dataclass(frozen=True)
class AdjClose(BaseModel):

    adjclose: List[float] = field(default_factory=list)


class Indicators(BaseModel):

    quote: Quote
    adjclose: AdjClose

    def __init__(self, indicators: IndicatorsDict) -> None:
        self.quote = Quote(**indicators.get("quote", [{}])[0])
        self.adjclose = AdjClose(**indicators.get("adjclose", [{}])[0])


@dataclass(frozen=True)
class TradingPeriod(OptionalFieldsModel):

    timezone: str
    start: int
    end: int
    gmtoffset: int


class CurrentTradingPeriod(BaseModel):

    pre: TradingPeriod
    regular: TradingPeriod
    post: TradingPeriod
    tradingPeriods: Optional[List[TradingPeriod]]

    def __init__(
        self,
        pre: TradingPeriodDict,
        regular: TradingPeriodDict,
        post: TradingPeriodDict,
        tradingPeriods: Optional[List[List[TradingPeriodDict]]] = None,
    ) -> None:
        self.pre = TradingPeriod(**pre)
        self.regular = TradingPeriod(**regular)
        self.post = TradingPeriod(**post)
        if tradingPeriods is not None:
            self.tradingPeriods = [TradingPeriod(**period) for periods in tradingPeriods for period in periods]


class Range(enum.Enum):

    NONE = ""
    ONE_DAY = "1d"
    FIVE_DAYS = "5d"
    ONE_MONTH = "1mo"
    THREE_MONTHS = "3mo"
    SIX_MONTHS = "6mo"
    ONE_YEAR = "1y"
    TWO_YEARS = "2y"
    FIVE_YEARS = "5y"
    TEN_YEARS = "10y"
    YTD = "ytd"
    MAX = "max"


class ChartMeta(BaseModel):

    currency: str
    symbol: str
    exchangeName: str
    instrumentType: str
    firstTradeDate: int
    regularMarketTime: int
    gmtoffset: int
    timezone: str
    exchangeTimezoneName: str
    regularMarketPrice: float
    chartPreviousClose: float
    priceHint: int
    currentTradingPeriod: CurrentTradingPeriod
    dataGranularity: Interval
    range: Range
    validRanges: List[Range]
    previousClose: Optional[float]
    scale: Optional[int]
    tradingPeriods: Optional[List[TradingPeriod]]

    def __init__(
        self,
        currency: str,
        symbol: str,
        exchangeName: str,
        instrumentType: str,
        firstTradeDate: int,
        regularMarketTime: int,
        gmtoffset: int,
        timezone: str,
        exchangeTimezoneName: str,
        regularMarketPrice: float,
        chartPreviousClose: float,
        priceHint: int,
        currentTradingPeriod: CurrentTradingPeriodDict,
        dataGranularity: str,
        range: Optional[str],
        validRanges: List[str],
        previousClose: Optional[float] = None,
        scale: Optional[int] = None,
        tradingPeriods: Optional[List[List[TradingPeriodDict]]] = None,
    ) -> None:
        self.currency = currency
        self.symbol = symbol
        self.exchangeName = exchangeName
        self.instrumentType = instrumentType
        self.firstTradeDate = firstTradeDate
        self.regularMarketTime = regularMarketTime
        self.gmtoffset = gmtoffset
        self.timezone = timezone
        self.exchangeTimezoneName = exchangeTimezoneName
        self.regularMarketPrice = regularMarketPrice
        self.chartPreviousClose = chartPreviousClose
        self.priceHint = priceHint
        self.currentTradingPeriod = CurrentTradingPeriod(**currentTradingPeriod)
        self.dataGranularity = Interval(dataGranularity)
        self.range = Range(range)
        self.validRanges = [Range(valid_range) for valid_range in validRanges]
        self.previousClose = previousClose
        self.scale = scale
        self.tradingPeriods = [
            TradingPeriod(**trading_period)
            for trading_period_sequence in tradingPeriods or [[]]
            for trading_period in trading_period_sequence
        ]

    def to_dict(self) -> ChartMetaDictBase:
        return {
            "currency": self.currency,
            "symbol": self.symbol,
            "exchangeName": self.exchangeName,
            "instrumentType": self.instrumentType,
            "firstTradeDate": self.firstTradeDate,
            "regularMarketTime": self.regularMarketTime,
            "gmtoffset": self.gmtoffset,
            "timezone": self.timezone,
            "exchangeTimezoneName": self.exchangeTimezoneName,
            "regularMarketPrice": self.regularMarketPrice,
            "chartPreviousClose": self.chartPreviousClose,
            "previousClose": self.previousClose,
            "priceHint": self.priceHint,
            "dataGranularity": self.dataGranularity.value,
            "range": self.range.value or None,
            "scale": self.scale,
        }



from typing import List, Optional, TypedDict


class QuoteDict(TypedDict, total=False):

    high: List[float]
    volume: List[float]
    open: List[float]
    close: List[float]
    low: List[float]


class AdjCloseDict(TypedDict, total=False):

    adjclose: List[float]


class IndicatorsDict(TypedDict):

    quote: List[QuoteDict]
    adjclose: List[AdjCloseDict]


class TradingPeriodDict(TypedDict):

    timezone: str
    start: int
    end: int
    gmtoffset: int


class CurrentTradingPeriodDict(TypedDict):

    pre: TradingPeriodDict
    regular: TradingPeriodDict
    post: TradingPeriodDict


class ChartMetaDictBase(TypedDict):

    currency: str
    symbol: str
    exchangeName: str
    instrumentType: str
    firstTradeDate: int
    regularMarketTime: int
    gmtoffset: int
    timezone: str
    exchangeTimezoneName: str
    regularMarketPrice: float
    chartPreviousClose: float
    priceHint: int
    dataGranularity: str
    range: Optional[str]
    scale: Optional[int]
    previousClose: Optional[float]


class ChartMetaDict(ChartMetaDictBase, total=False):

    currentTradingPeriod: CurrentTradingPeriodDict
    validRanges: List[str]
    tradingPeriods: List[List[TradingPeriodDict]]


class ChartDataRecord(ChartMetaDictBase):

    timestamp: int
    high: float
    low: float
    volume: float
    open: float
    close: float
    adjclose: float



import asyncio
from typing import Any, Dict, Iterable, List, cast

import aiohttp

from pyhoo.getter import GetTickerDataTask
from pyhoo.types import ApiResponse


class Requester:

    _path: str
    _tickers: Iterable[str]
    _max_concurrent_calls: int
    _params: Dict[str, Any]

    def __init__(
        self,
        path: str,
        tickers: Iterable[str],
        max_concurrent_calls: int = 100,
        **params: Any,
    ) -> None:
        self._path = path
        self._tickers = tickers
        self._params = params
        self._max_concurrent_calls = max_concurrent_calls

    async def request(self) -> List[Dict[str, ApiResponse]]:
        """Asynchronously fire requests by ticker thanks to the `GetTickerDataTask`."""
        connector = aiohttp.TCPConnector(limit=self._max_concurrent_calls)
        async with aiohttp.ClientSession(connector=connector) as session:
            tasks = []
            for ticker in self._tickers:
                tasks.append(
                    GetTickerDataTask(
                        path=self._path,
                        ticker=ticker,
                        **self._params,
                    ).run(session=session)
                )
            responses = cast(
                List[Dict[str, ApiResponse]],
                await asyncio.gather(*tasks, return_exceptions=False),
            )
            return responses



from dataclasses import dataclass
from typing import Optional, Sequence, TypeVar

from pyhoo.models.abc import BaseModel, OptionalFieldsModel
from pyhoo.types.fundamentals import ReportedValueDict


@dataclass(frozen=True)
class ReportedValue(OptionalFieldsModel):

    raw: float
    fmt: str


class FundamentalsRow(BaseModel):

    dataId: int
    asOfDate: str
    periodType: str
    reportedValue: ReportedValue
    currencyCode: Optional[str]

    def __init__(
        self,
        dataId: int,
        asOfDate: str,
        periodType: str,
        reportedValue: ReportedValueDict,
        currencyCode: Optional[str] = None,
    ) -> None:
        self.dataId = dataId
        self.asOfDate = asOfDate
        self.periodType = periodType
        self.reportedValue = ReportedValue(**reportedValue)
        self.currencyCode = currencyCode


_T = TypeVar("_T")


class FundamentalsMeta(BaseModel):

    symbol: str
    type: str

    def __init__(self, symbol: Sequence[str], type: Sequence[str]) -> None:
        self.symbol = self._get_first_element(symbol)
        self.type = self._get_first_element(type)

    @staticmethod
    def _get_first_element(sequence: Sequence[_T]) -> _T:
        return sequence[0]



import datetime
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)

from pyhoo.errors import (
    InvalidParameterPrefixError,
    InvalidParameterTypeError,
    InvalidParameterValueError,
    MissingParameterError,
    UnknownParameterError,
)
from pyhoo.parsers import ChartParser, FundamentalsParser, OptionsParser
from pyhoo.parsers.abc import BaseParser

_T = TypeVar("_T")
_V = TypeVar("_V")

FUNDAMENTALS_TYPE_OPTIONS_PATH = Path(__file__).parent / "data/fundamentals_type_options.txt"


class ParamConfig:
    def __init__(
        self,
        name: str,
        api_name: str,
        type: Type[_T],
        required: bool = False,
        converter: Optional[Callable[[_T], _V]] = None,
        default: Optional[Any] = None,
        options: Optional[Iterable[_T]] = None,
        prefixes: Optional[Iterable[str]] = None,
    ) -> None:
        self.name = name
        self.api_name = api_name
        self.type = type
        self.converter = converter
        self.default = default
        self.required = required
        self.options = {option for option in options or []}
        self.prefixes = prefixes or []

    def validate(self, value: _T) -> None:
        if not isinstance(value, self.type):
            raise InvalidParameterTypeError(self.name, type(value), self.type)
        if self.options:
            values = cast(Iterable[_T], [value] if self.type != list else value)
            for _value in values:
                unprefixed_value = self._unprefix(cast(str, _value))
                if unprefixed_value not in self.options:
                    raise InvalidParameterValueError(self.name, unprefixed_value, self.options)

    def format(self, value: _T) -> Union[_T, _V]:
        if self.converter:
            return self.converter(value)
        return value

    def _unprefix(self, value: str) -> Optional[str]:
        """Remove the longest found prefix (in `self.prefixes`) from value.
        Currently used to check fundamentals types options.
        """
        if self.prefixes:
            # We need to match the 'un-prefixed' value against the allowed options
            prefix_lengths = [len(prefix) for prefix in self.prefixes if value.startswith(prefix)]
            if prefix_lengths:
                # We match the longest prefix
                return value[max(prefix_lengths) :]
            else:
                raise InvalidParameterPrefixError(self.name, value, self.prefixes)
        return value


class Config:
    def __init__(
        self,
        path: str,
        response_field: str,
        parser: Type[BaseParser],
        params_config: Iterable[ParamConfig],
    ) -> None:
        self.path = path
        self.response_field = response_field
        self.parser = parser
        self.params_config = {param.name: param for param in params_config}

    def validate(self, params: Dict[str, Any]) -> None:
        for param, value in params.items():
            param_config = self.params_config.get(param)
            if param_config is None:
                raise UnknownParameterError(param, self.params_config)
            param_config.validate(value)
        for param, param_config in self.params_config.items():
            if param_config.required and param not in params:
                if param_config.default is not None:
                    params[param] = param_config.default
                else:
                    raise MissingParameterError(param)

    def format(self, params: Dict[str, Any]) -> Dict[str, Any]:
        return {
            self.params_config[param].api_name: self.params_config[param].format(value)
            for param, value in params.items()
        }


def str_date_to_timestamp(str_date: str) -> int:
    return int(datetime.datetime.timestamp(datetime.datetime.strptime(str_date, "%Y-%m-%d")))


with open(FUNDAMENTALS_TYPE_OPTIONS_PATH, "r") as file:
    fundamentals_type_options = [line.strip() for line in file.readlines()]


endpoints_config = {
    "chart": Config(
        path="v8/finance/chart",
        response_field="chart",
        parser=ChartParser,
        params_config=[
            ParamConfig(
                name="start",
                api_name="period1",
                type=str,
                required=True,
                converter=str_date_to_timestamp,
            ),
            ParamConfig(
                name="end",
                api_name="period2",
                type=str,
                required=True,
                converter=str_date_to_timestamp,
            ),
            ParamConfig(
                name="granularity",
                api_name="interval",
                type=str,
                required=True,
                default="1d",
                options=["1m", "2m", "5m", "15m", "30m", "1h", "1d", "5d", "1w", "1mo", "3mo"],
            ),
            ParamConfig(
                name="range",
                api_name="range",
                type=str,
                required=False,
                options=["1d", "5d", "1mo", "3mo", "6mo", "1y", "2y", "5y", "10y", "ytd", "max"],
            ),
        ],
    ),
    "fundamentals": Config(
        path="ws/fundamentals-timeseries/v1/finance/timeseries",
        response_field="timeseries",
        parser=FundamentalsParser,
        params_config=[
            ParamConfig(
                name="start",
                api_name="period1",
                type=str,
                required=False,
                converter=str_date_to_timestamp,
            ),
            ParamConfig(
                name="end",
                api_name="period2",
                type=str,
                required=False,
                converter=str_date_to_timestamp,
            ),
            ParamConfig(
                name="type",
                api_name="type",
                type=list,
                required=True,
                options=fundamentals_type_options,
                default=[f"annual{value}" for value in fundamentals_type_options],
                converter=lambda values: ",".join(values),
                prefixes=["monthly", "quarterly", "annual"],
            ),
        ],
    ),
    "options": Config(
        path="v7/finance/options",
        response_field="optionChain",
        parser=OptionsParser,
        params_config=[
            ParamConfig(
                name="start",
                api_name="date",
                type=str,
                required=False,
                converter=str_date_to_timestamp,
            ),
            ParamConfig(
                name="end",
                api_name="endDate",
                type=str,
                required=False,
                converter=str_date_to_timestamp,
            ),
            ParamConfig(
                name="strikeMin",
                api_name="strikeMin",
                type=float,
                required=False,
            ),
            ParamConfig(
                name="strikeMax",
                api_name="strikeMax",
                type=float,
                required=False,
            ),
        ],
    ),
}



from itertools import zip_longest
from typing import List, cast

from pyhoo.models.chart import ChartMeta, Indicators
from pyhoo.models.iterables import Timestamp
from pyhoo.parsers.abc import BaseParser
from pyhoo.types.chart import ChartDataRecord, ChartMetaDict, IndicatorsDict


class ChartParser(BaseParser):

    meta: ChartMeta
    timestamp: Timestamp
    indicators: Indicators

    def __init__(self, meta: ChartMetaDict, timestamp: List[int], indicators: IndicatorsDict) -> None:
        self.meta = ChartMeta(**meta)
        self.timestamp = Timestamp(timestamp)
        self.indicators = Indicators(indicators)

    def to_records(self) -> List[ChartDataRecord]:
        return [
            cast(
                ChartDataRecord,
                {
                    "timestamp": timestamp,
                    "high": high,
                    "low": low,
                    "volume": volume,
                    "open": open,
                    "close": close,
                    "adjclose": adjclose,
                    **self.meta.to_dict(),
                },
            )
            for timestamp, high, low, volume, open, close, adjclose in zip_longest(
                self.timestamp,
                self.indicators.quote.high,
                self.indicators.quote.low,
                self.indicators.quote.volume,
                self.indicators.quote.open,
                self.indicators.quote.close,
                self.indicators.adjclose.adjclose,
            )
        ]



from typing import Any, Dict, List, Literal, Optional, TypedDict


class ErrorDescription(TypedDict):
    code: str
    description: str


class ApiResponse(TypedDict):
    result: Optional[List[Dict[str, List[Any]]]]
    error: Optional[ErrorDescription]


Endpoint = Literal["chart", "fundamentals", "options"]



import asyncio
from typing import Any, Dict, Iterable, Type, Union, cast

import pandas as pd

from pyhoo.config import endpoints_config
from pyhoo.errors import ApiError
from pyhoo.parsers.abc import BaseParser
from pyhoo.requester import Requester
from pyhoo.types import ApiResponse, Endpoint, ErrorDescription


def get(
    endpoint: Endpoint,
    tickers: Union[str, Iterable[str]],
    max_concurrent_calls: int = 100,
    ignore_errors: bool = False,
    **params: Any,
) -> pd.DataFrame:
    if not _is_iterable(tickers):
        tickers = [cast(str, tickers)]
    endpoint_config = endpoints_config[endpoint]
    endpoint_config.validate(params)
    responses = asyncio.run(
        Requester(
            path=endpoint_config.path,
            tickers=tickers,
            max_concurrent_calls=max_concurrent_calls,
            **endpoint_config.format(params),
        ).request()
    )
    return _convert_to_dataframe(
        responses=responses,
        response_field=endpoint_config.response_field,
        parser=endpoint_config.parser,
        ignore_errors=ignore_errors,
    )


def _is_iterable(obj: Any) -> bool:
    """Check that an object is iterable but not a string (strings are iterable)."""
    return hasattr(obj, "__iter__") and not isinstance(obj, str)


def _convert_to_dataframe(
    responses: Iterable[Dict[str, ApiResponse]],
    response_field: str,
    parser: Type[BaseParser],
    ignore_errors: bool,
) -> pd.DataFrame:
    records = []
    for response in responses:
        response_data = response[response_field]
        if response_data.get("error") is not None:
            if ignore_errors:
                continue
            error = cast(ErrorDescription, response_data["error"])
            raise ApiError(error["code"], error["description"])
        result = response_data["result"]
        if result is not None:
            records += [record for data in result for record in parser(**data).to_records()]
    return pd.DataFrame(records)



"""
    Define a `CustomIterable[T]` generic type, to be used as an `attr.ib()`
    to parse sequences of any type with:
    `<ATTR_NAME>: CustomIterable = attr.ib(converter=CustomIterable)`.

    Ex:
        # Timestamp is defined below
        timestamp: Timestamp = attr.ib(converter=Timestamp)
"""
from __future__ import annotations

import datetime
from typing import Generic, List, TypeVar

from pyhoo.models.abc import BaseModel

_T = TypeVar("_T")


class CustomIterable(BaseModel, Generic[_T]):

    _values: List[_T]

    def __init__(self, values: List[_T]) -> None:
        self._values = values

    def __iter__(self) -> "CustomIterator[_T]":
        return CustomIterator(self)

    def __getitem__(self, index: int) -> _T:
        return self._values[index]

    def __len__(self) -> int:
        return len(self._values)


class CustomIterator(Generic[_T]):

    _iterable: CustomIterable[_T]
    _index: int

    def __init__(self, iterable: CustomIterable[_T]) -> None:
        self._iterable = iterable
        self._index = 0

    def __next__(self) -> _T:
        if self._index < len(self._iterable):
            result = self._iterable[self._index]
            self._index += 1
            return result
        raise StopIteration

    def __iter__(self) -> CustomIterator[_T]:
        return self


class Timestamp(CustomIterable[int]):
    def to_str(self, format: str = "%Y-%m-%dT%H:%M:%S") -> List[str]:
        """Convert each timestamp to a datetime object and then format it as specified."""
        return list(map(lambda date: date.strftime(format), self.to_datetime()))

    def to_datetime(self) -> List[datetime.datetime]:
        """Convert each timestamp to a datetime object."""
        return list(map(datetime.datetime.fromtimestamp, self))


class Strikes(CustomIterable[float]):
    pass



from dataclasses import asdict
from typing import List, Sequence, cast

from pyhoo.models.iterables import Strikes, Timestamp
from pyhoo.models.options import OptionQuote, Options
from pyhoo.parsers.abc import BaseParser
from pyhoo.types.options import OptionQuoteDict, OptionsDataRecord, OptionsDict


class OptionsParser(BaseParser):

    underlyingSymbol: str
    expirationDates: Timestamp
    strikes: Strikes
    hasMiniOptions: bool
    quote: OptionQuote
    options: Options

    def __init__(
        self,
        underlyingSymbol: str,
        expirationDates: List[int],
        strikes: List[float],
        hasMiniOptions: bool,
        quote: OptionQuoteDict,
        options: Sequence[OptionsDict] = None,
    ) -> None:
        self.underlyingSymbol = underlyingSymbol
        self.expirationDates = Timestamp(expirationDates)
        self.strikes = Strikes(strikes)
        self.hasMiniOptions = hasMiniOptions
        self.quote = OptionQuote(**quote)
        self.options = Options(**options[0]) if options else Options()

    def to_records(self) -> List[OptionsDataRecord]:
        return [
            cast(
                OptionsDataRecord,
                {
                    "underlyingSymbol": self.underlyingSymbol,
                    "type": type,
                    **asdict(option),
                },
            )
            for option, type in [(call, "CALL") for call in self.options.calls]
            + [(put, "PUT") for put in self.options.puts]
        ]



from typing import Dict, Iterable, List, Optional

from pyhoo.models.fundamentals import FundamentalsMeta, FundamentalsRow
from pyhoo.models.iterables import Timestamp
from pyhoo.parsers.abc import BaseParser
from pyhoo.types.fundamentals import (
    FundamentalsDataRowDict,
    FundamentalsMetaDict,
    FundamentalsRowDict,
)


class FundamentalsParser(BaseParser):
    def __init__(
        self,
        meta: FundamentalsMetaDict,
        timestamp: Optional[List[int]] = None,
        **data: Iterable[FundamentalsRowDict],
    ) -> None:
        self.meta = FundamentalsMeta(**meta)
        self.timestamp = Timestamp(timestamp or [])
        self._parse_data(data)

    def _parse_data(self, data: Dict[str, Iterable[FundamentalsRowDict]]) -> None:
        fundamentals_name = self.meta.type
        fundamentals_data = next(iter(data.values())) if data else []
        setattr(
            self,
            fundamentals_name,
            [FundamentalsRow(**row) for row in fundamentals_data],
        )

    def to_records(self) -> List[FundamentalsDataRowDict]:
        return [
            {
                "type": self.meta.type,
                "symbol": self.meta.symbol,
                "dataId": row.dataId,
                "asOfDate": row.asOfDate,
                "periodType": row.periodType,
                "reportedValue": row.reportedValue.raw,
                "currencyCode": row.currencyCode,
            }
            for row in getattr(self, self.meta.type)
        ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyhoo', 'pyhoo.models', 'pyhoo.parsers', 'pyhoo.types']

package_data = \
{'': ['*'], 'pyhoo': ['data/*']}

install_requires = \
['aiohttp>=3.6.1,<4.0.0', 'pandas>=1.0.0,<2.0.0']

setup_kwargs = {
    'name': 'pyhoo',
    'version': '0.1.3',
    'description': 'Yet another unofficial Yahoo Finance API, but with concurrent requests',
    'long_description': '# Pyhoo\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![Build Status](https://travis-ci.com/prouhard/pyhoo.svg?branch=master)](https://travis-ci.com/prouhard/pyhoo)\n[![codecov](https://codecov.io/gh/prouhard/pyhoo/branch/master/graph/badge.svg?token=6VJW1F01DL)](https://codecov.io/gh/prouhard/pyhoo)\n\n_Yet another unofficial Yahoo Finance API library but with concurrent requests._\n\n**Index**\n\n1. [Installation](#installation)\n1. [Usage](#usage)\n1. [Troubleshooting](#troubleshooting)\n1. [Contributing](#contributing)\n\n## Installation\n\n**Pyhoo requires Python >= 3.8**\n\n```bash\npip install pyhoo\n```\n\n## Usage\n\nPyhoo is **simple**:\n\n```python\nimport pyhoo\n\ntickers = [\'FB\', \'AAPL\', \'AMZN\', \'GOOGL\']\nstart = \'2020-02-01\'\nend = \'2020-11-02\'\n\nstock_prices = pyhoo.get(\'chart\', tickers, start=start, end=end, granularity="1d")\nfinancial_reports = pyhoo.get(\'fundamentals\', tickers, start=start, end=end)\noptions = pyhoo.get(\'options\', tickers, strikeMax=400.0)\n```\n\nThe result of `pyhoo.get` is a formatted `pandas.DataFrame` (here for stock prices):\n\n|     |  timestamp |   high |    low |   volume |   open |  close | adjclose | currency | symbol | exchangeName | instrumentType | regularMarketPrice | ... |\n| --: | ---------: | -----: | -----: | -------: | -----: | -----: | -------: | :------- | :----- | :----------- | :------------- | -----------------: | --: |\n|   0 | 1580481000 | 208.69 | 201.06 | 31359900 | 208.43 | 201.91 |   201.91 | USD      | FB     | NMS          | EQUITY         |             286.95 | ... |\n|   1 | 1580740200 | 205.14 |  202.5 | 15510500 | 203.44 | 204.19 |   204.19 | USD      | FB     | NMS          | EQUITY         |             286.95 | ... |\n|   2 | 1580826600 |  210.6 |  205.2 | 19628900 | 206.62 | 209.83 |   209.83 | USD      | FB     | NMS          | EQUITY         |             286.95 | ... |\n|   3 | 1580913000 | 212.73 | 208.71 | 12538200 | 212.51 | 210.11 |   210.11 | USD      | FB     | NMS          | EQUITY         |             286.95 | ... |\n|   4 | 1580999400 | 211.19 | 209.34 | 10567500 | 210.47 | 210.85 |   210.85 | USD      | FB     | NMS          | EQUITY         |             286.95 | ... |\n|   5 |        ... |    ... |    ... |      ... |    ... |    ... |      ... | ...      | ...    | ...          | ...            |                ... | ... |\n\nPyhoo is **fast**, it uses concurrency to fire multiple requests at the same time. You can request all the tickers of the S&P500 in one shot.\n\nPyhoo is **still in development**, feel free to add more endpoints thanks to the `Config` object !\n\nCurrently, it supports three endpoints:\n\n1. `chart`, for [OHLC](https://en.wikipedia.org/wiki/Open-high-low-close_chart) data, basically stock prices\n1. `fundamentals`, for financial data about the firm, see the [list of available reports](pyhoo/data/fundamentals_type_options.txt)\n1. `options`, for detailed information on each call and put at each strike on specific tickers\n\n## Troubleshooting\n\nIf running from a Jupyter Notebook, you may encounter the following error:\n\n```python\nRuntimeError: asyncio.run() cannot be called from a running event loop\n```\n\nThis is because Jupyter Notebooks are running themselves in an event loop, and it is a known issue with `asyncio.run`.\n\nThere is a workaround, a bit hacky but gets the job done, using [nest_asyncio](https://github.com/erdewit/nest_asyncio).\n\n```bash\npip install nest_asyncio\n```\n\nThen in the Notebook, before calling `pyhoo.get`:\n\n```python\nimport nest_asyncio\nnest_asyncio.apply()\n```\n\nAnd you should be ok !\n\n## Contributing\n\nContributions are welcome !\n\nPyhoo uses [poetry](https://python-poetry.org) as package manager. You can find the installation instructions [here](https://python-poetry.org/docs/#installation).\n\nIt is recommended to install the virtual environment in the project folder if you use VSCode, to help the linter resolve imports:\n\n```bash\npoetry config virtualenvs.path --unset\npoetry config virtualenvs.in-project true\n```\n\nOnce Poetry is configured, you can install required dependencies with:\n\n```bash\npoetry install\n```\n\nThe CI enforces strict typing, linting and coverage.\n',
    'author': 'ROUHARD Pierre',
    'author_email': 'pierrerouhard@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import abc
from typing import Any, List


class BaseParser(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    @abc.abstractmethod
    def to_records(self) -> List[Any]:
        pass

    def __repr__(self) -> str:
        formatted_attrs = ", ".join(
            attr_name + "=" + attr_value.__repr__() for attr_name, attr_value in self.__dict__.items()
        )
        return f"{self.__class__.__name__}({formatted_attrs})"
